module jp.ku6.windows.game.animation;

import std.stream;
import jp.ku6.windows.win32api.windows;
import jp.ku6.windows.directx.dxgiformat;
import jp.ku6.windows.directx.dxgitype;
import jp.ku6.windows.directx.dxgi;
import jp.ku6.windows.directx.d3dcommon;
import jp.ku6.windows.directx.d3d11;
import jp.ku6.windows.directx.d3dcompiler;
import jp.ku6.windows.game.graphics;

/++
 + 接触判定用矩形
 +/
struct CollisionRect
{
	/// 上端
	float top;

	/// 下端
	float bottom;

	/// 左端
	float left;

	/// 右端
	float right;

	/++
	 + 接触判定
	 +
	 + Params:
	 +  dx = 矩形の左上点X座標
	 +  dy = 矩形の左上点Y座標
	 +  sx = 判定対象矩形の左上点X座標
	 +  sy = 判定対象矩形の左上点Y座標
	 +  src = 判定対象矩形
	 +
	 + Returns:
	 +  接触している場合にtrueを返す
	 +/
	const
	bool isCollision(float dx, float dy, float sx, float sy, CollisionRect* src)
	nothrow
	{
		return src !is null && top + dy <= src.bottom + sy && bottom + dy >= src.top + sy && left + dx <= src.right + sx && right + dx >= src.left + sx;
	}
}

/++
 + 一定時間内の接触判定用データ
 +/
struct CollisionFrame
{
	// 使用終了時間
	ushort time;

	// 矩形の数
	ushort length;

	// 矩形データ
	CollisionRect* rects;

	/++
	 + 接触判定
	 +
	 + Params:
	 +  dx = X座標
	 +  dy = Y座標
	 +  sx = 判定対象のX座標
	 +  sy = 判定対象のY座標
	 +  src = 判定対象
	 +
	 + Returns:
	 +  接触している場合にtrueを返す
	 +/
	const
	bool isCollision(float dx, float dy, float sx, float sy, CollisionFrame* src)
	nothrow
	{
		for (ushort d; d < length; ++d)
		{
			for (ushort s; s < src.length; ++s)
			{
				if (rects[d].isCollision(dx, dy, sx, sy, &src.rects[s]))
					return true;
			}
		}
		return false;
	}
}

/++
 + 接触判定用データ
 +/
struct CollisionTimeline
{
	/// 接触判定用フレームの個数
	uint length;

	/// 接触判定用フレーム
	CollisionFrame* frames;
}

/++
 + アニメーション中で画像を組み合わせるためのパーツ
 +/
struct AnimationPart
{
	/// 表示タイプ
	uint type;

	/// パターン
	Pattern* pattern;

	/// 色
	uint color;

	/// 8byte alignのためのダミー
	uint dummy;

	/// 4頂点情報
	float[8] vertices;

	/++
	 + 描画
	 +
	 + Params:
	 +  x = 描画位置X座標
	 +  y = 描画位置Y座標
	 +/
	void draw(float x, float y)
	{
		if (pattern is null || pattern.surface is null || pattern.surface.textureResourceView is null)
			return;

		// Spriteレンダリング設定
		AnimationPartRenderer.Constants* constants = &Animation.renderer_.constants_;
		constants.x = x;
		constants.y = y;
		constants.u = pattern.u;
		constants.v = pattern.v;
		constants.width = pattern.width;
		constants.height = pattern.height;
		constants.textureWidth = pattern.surface.width;
		constants.textureHeight = pattern.surface.height;
		constants.color = color;
		constants.xarray[0] = vertices[0];
		constants.xarray[1] = vertices[2];
		constants.xarray[2] = vertices[6];
		constants.xarray[3] = vertices[4];
		constants.yarray[0] = vertices[1];
		constants.yarray[1] = vertices[3];
		constants.yarray[2] = vertices[7];
		constants.yarray[3] = vertices[5];

		// 描画処理
		Animation.renderer_.render(Graphics.instance.immediateContext, pattern.surface.textureResourceView);
	}
}

/++
 + アニメーション中で画像パーツを組み合わせたもの
 +/
struct AnimationFrame
{
	/// 描画時間
	ushort time;

	/// パーツ数
	ushort length;

	/// パーツ
	AnimationPart* parts;

	/++
	 + 描画
	 +
	 + Params:
	 +  x = 描画位置X座標
	 +  y = 描画位置Y座標
	 +/
	void draw(float x, float y)
	{
		for (ushort i; i < length; ++i)
			parts[i].draw(x, y);
	}
}

/++
 + アニメーションデータ
 +/
struct AnimationTimeline
{
	// 描画フレーム数
	uint length;

	// 接触判定数
	uint collisionTimelineLength;

	// 描画フレーム
	AnimationFrame* animationFrames;

	// 接触判定
	CollisionTimeline* collisionTimelines;
}

/++
 + ファイル内の圧縮データをサーフェスに転送するクラス
 +/
class MemoryImage : SurfaceSource
{
	/++
	 + データをセット
	 +
	 + Params:
	 +  buffer = データポインタ
	 +  width = 画像の幅
	 +  height = 画像の高さ
	 +/
	void set(uint* buffer, uint width, uint height)
	{
		width_ = width;
		height_ = height;
		buffer_ = buffer;
	}

	uint width()
	{
		return width_;
	}

	uint height()
	{
		return height_;
	}

	string filePath()
	{
		return null;
	}

	bool open()
	{
		return true;
	}

	void close()
	{
	}

	void read(uint* buffer, uint width, uint height)
	{
		for (uint y; y < height_; ++y)
		{
			uint* d = &buffer[y * width];
			uint* s = &buffer_[y * width_];
			for (uint x; x < width_; ++x)
			{
				// Direct3D 9向けのファイルだったので、RとBを反転させる
				uint a = s[x];
				d[x] = (a & 0xff00ff00) | ((a >> 16) & 0xff) | ((a << 16) & 0xff0000);
			}
		}
	}

private:

	// 画像の幅
	uint width_;

	// 画像の高さ
	uint height_;

	// データのポインタ
	uint* buffer_;
}

/++
 + アニメーションファイル
 +
 + ツールで作成したkagファイルをセットし、必要なデータを抜き出すためのインターフェース
 +/
class AnimationFile
{
	/++
	 + コンストラクタ
	 +/
	this()
	{
	}

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +/
	this(string filePath, uint offset = 0)
	{
		set(filePath, offset);
	}

	/++
	 + デストラクタ
	 +/
	~this()
	{
		release();
	}

	/++
	 + 内部で保持しているデータをクリアする
	 +/
	void release()
	{
		foreach (s; surfaceArray_)
			delete s;

		surfaceArray_ = null;
		patternArray_ = null;
		animationPartArray_ = null;
		animationFrameArray_ = null;
		animationTimelineArray_ = null;
		collisionRectArray_ = null;
		collisionFrameArray_ = null;
		collisionTimelineArray_ = null;
		surfaces_ = null;
		patterns_ = null;
		animationTimelines_ = null;
	}

	/++
	 + サーフェスを取得
	 +
	 + Params:
	 +  number = 識別番号
	 +
	 + Returns:
	 +  サーフェス
	 +  識別番号が間違っていた場合にはnullを返す
	 +/
	Surface getSurface(uint number)
	{
		Surface* surface = number in surfaces_;
		if (surface is null)
			return null;
		else
			return *surface;
	}

	/++
	 + パターンを取得
	 +
	 + Params:
	 +  number = 識別番号
	 +
	 + Returns:
	 +  パターン
	 +  識別番号が間違っていた場合にはnullを返す
	 +/
	Pattern* getPattern(uint number)
	{
		Pattern** pattern = number in patterns_;
		if (pattern is null)
			return null;
		else
			return *pattern;
	}

	/++
	 + アニメーションデータを取得
	 +
	 + Params:
	 +  number = 識別番号
	 +
	 + Returns:
	 +  アニメーションデータ
	 +  識別番号が間違っていた場合にはnullを返す
	 +/
	AnimationTimeline* getAnimationTimeline(uint number)
	{
		AnimationTimeline** timeline = number in animationTimelines_;
		if (timeline is null)
			return null;
		else
			return *timeline;
	}

	/++
	 + ファイルをセット
	 +
	 + Params:
	 +  filePath = ファイルパス
	 +/
	void set(string filePath, uint offset = 0)
	{
		// 現在のデータは消去
		release();

		// ファイルをオープンする
		auto file = new BufferedFile(filePath, FileMode.In);
		scope(exit) file.close();

		if (offset != 0)
			file.seekSet(offset);

		// ファイルヘッダを読み込み
		FileHeader header;
		file.readExact(&header, FileHeader.sizeof);


		// 圧縮画像読み込み用バッファ
		ubyte[] sourceBuffer = new ubyte[header.compressedImageMax];

		// 解凍画像書き込み用バッファ
		ubyte[] destinationBuffer = new ubyte[header.uncompressedImageMax];

		// 画像のヘッダ情報を取得
		SurfaceHeader[] surfaceHeaders = new SurfaceHeader[header.surfaceLength];
		file.read(cast(ubyte[])surfaceHeaders);

		// 画像の識別番号を読み込み
		uint[] surfaceTags = new uint[header.surfaceLength];
		file.read(cast(ubyte[])surfaceTags);

		// サーフェスを作成
		auto image = new MemoryImage;
		surfaceArray_ = new Surface[header.surfaceLength];
		for (uint i; i < header.surfaceLength; ++i)
		{
			// ピクセルデータをメモリ上にセット
			file.readExact(sourceBuffer.ptr, surfaceHeaders[i].length);
			uncompress(sourceBuffer.ptr, surfaceHeaders[i].length, destinationBuffer.ptr);
			image.set(cast(uint*)destinationBuffer.ptr, surfaceHeaders[i].width, surfaceHeaders[i].height);

			// テクスチャを作成してセット
			surfaceArray_[i] = new Surface;
			surfaceArray_[i].load(image);
			surfaces_[surfaceTags[i]] = surfaceArray_[i];
		}

		// パターンの識別番号を読み込み
		uint[] patternTags = new uint[header.patternLength];
		file.read(cast(ubyte[])patternTags);

		// パターンのロード
		int[] patternBytes = new int[header.patternLength * 5];
		file.read(cast(ubyte[])patternBytes);

		patternArray_ = new Pattern[header.patternLength];
		for (uint i; i < header.patternLength; ++i)
		{
			patternArray_[i].u = patternBytes[i * 5];
			patternArray_[i].v = patternBytes[i * 5 + 1];
			patternArray_[i].width = patternBytes[i * 5 + 2];
			patternArray_[i].height = patternBytes[i * 5 + 3];
			patternArray_[i].surface = surfaceArray_[patternBytes[i * 5 + 4]];

			patterns_[patternTags[i]] = &patternArray_[i];
		}

		// あたり判定はある？
		if ((header.flags & 0x00000002) != 0)
		{
			// 矩形範囲のバッファを作成
			collisionRectArray_ = new CollisionRect[header.collisionRectLength];
			file.read(cast(ubyte[])collisionRectArray_);

			// あたり判定のバッファを作成
			collisionFrameArray_ = new CollisionFrame[header.collisionFrameLength];
			file.read(cast(ubyte[])collisionFrameArray_);

			// あたり判定データのバッファを作成
			collisionTimelineArray_ = new CollisionTimeline[header.collisionTimelineLength];
			file.read(cast(ubyte[])collisionTimelineArray_);
		}

		// アニメーションはある？
		if ((header.flags & 0x00000004) != 0)
		{
			// アニメーションパーツの個数を取得し、バッファを作成
			animationPartArray_ = new AnimationPart[header.animationPartLength];
			file.read(cast(ubyte[])animationPartArray_);

			// アニメーションフレームの個数を取得し、バッファを作成
			animationFrameArray_ = new AnimationFrame[header.animationFrameLength];
			file.read(cast(ubyte[])animationFrameArray_);

			// アニメーション識別番号領域を作成
			uint[] animationTagArray = new uint[header.animationTimelineLength];
			file.read(cast(ubyte[])animationTagArray);

			// アニメーションデータの個数を取得し、バッファ作成
			animationTimelineArray_ = new AnimationTimeline[header.animationTimelineLength];
			file.read(cast(ubyte[])animationTimelineArray_);

			for (uint i; i < animationTimelineArray_.length; ++i)
				animationTimelines_[animationTagArray[i]] = &animationTimelineArray_[i];
		}

		foreach (ref f; collisionFrameArray_)
			f.rects = &collisionRectArray_[cast(uint)(f.rects)];

		foreach (ref t; collisionTimelineArray_)
			t.frames = &collisionFrameArray_[cast(uint)(t.frames)];

		foreach (ref p; animationPartArray_)
			p.pattern = &patternArray_[cast(uint)(p.pattern)];

		foreach (ref f; animationFrameArray_)
			f.parts = &animationPartArray_[cast(uint)(f.parts)];

		foreach (ref t; animationTimelineArray_)
		{
			t.animationFrames = &animationFrameArray_[cast(uint)(t.animationFrames)];
			if (cast(uint)t.collisionTimelines > 0)
				t.collisionTimelines = &collisionTimelineArray_[cast(uint)(t.collisionTimelines)];
		}
	}

private:

	/+
	 + ファイルヘッダ
	 +/
	static
	struct FileHeader
	{
		// GA\0の文字
		char acTag[4];

		// 各種フラグ
		uint flags;

		// 内部で保持している圧縮済み画像の最大サイズ
		uint compressedImageMax;

		// 内部で保持している圧縮済み画像を展開した時の最大サイズ
		uint uncompressedImageMax;

		// テクスチャ数
		uint surfaceLength;

		// パターン数
		uint patternLength;

		// アニメーションパーツの数
		uint animationPartLength;

		// アニメーションフレームの数
		uint animationFrameLength;

		// アニメーションタイムラインの数
		uint animationTimelineLength;

		// 接触判定用矩形の数
		uint collisionRectLength;

		// 接触用データの数
		uint collisionFrameLength;

		// 接触用データのタイムラインの数
		uint collisionTimelineLength;
	}

	/+
	 + 
	 +/
	static
	struct SurfaceHeader
	{
		// 幅
		ushort width;

		// 高さ
		ushort height;

		// データ長
		uint length;
	}

	// サーフェスサーフェスの配列
	Surface[] surfaceArray_;

	// パターンの配列
	Pattern[] patternArray_;

	// アニメーションパーツの配列
	AnimationPart[] animationPartArray_;

	// アニメーションフレームの配列
	AnimationFrame[] animationFrameArray_;

	// アニメーションデータの配列
	AnimationTimeline[] animationTimelineArray_;

	// 接触判定矩形の配列
	CollisionRect[] collisionRectArray_;

	// 接触判定フレームの配列
	CollisionFrame[] collisionFrameArray_;

	// 接触判定用データの配列
	CollisionTimeline[] collisionTimelineArray_;

	// サーフェスデータへのマップ
	Surface[uint] surfaces_;

	// パターンデータへのマップ
	Pattern*[uint] patterns_;

	// アニメーションデータへのマップ
	AnimationTimeline*[uint] animationTimelines_;

	/+
	 + ピクセルデータの解凍処理
	 +
	 + Params:
	 +  source = 圧縮データのポインタ
	 +  sourceLength = 圧縮データの長さ
	 +  destination = 解凍後のデータの格納領域
	 +/
	void uncompress(ubyte* source, uint sourceLength, ubyte* destination)
	{
		uint sourceCursor;
		uint destinationCursor;

		uint flags;
		while (sourceCursor < sourceLength)
		{
			// 圧縮フラグを読みきったら、圧縮フラグを1バイト取得
			flags >>= 1;
			if ((flags & 0x100) == 0 )
			{
				flags = source[sourceCursor++] | 0x0000FF00;
			}

			// 圧縮データか？
			if ((flags & 1) == 0)
			{
				// 一致位置と一致長を取得
				uint matchPosition = cast(uint)source[sourceCursor++];
				uint matchLength = cast(uint)source[sourceCursor++];
				matchPosition |= ((matchLength & 0xF0) << 4);
				matchLength = (matchLength & ~0xF0) + 3;

				// コピー元のデータのポインタを取得
				ubyte* matchedBytes = &destination[destinationCursor - matchPosition];

				// 一致長分コピー
				for (uint i; i < matchLength; ++i)
					destination[destinationCursor++] = matchedBytes[i];
			}
			else
			{
				// ふつーに出力
				destination[destinationCursor++] = source[sourceCursor++];
			}
		}
	}
}

/++
 + アニメーション
 +/
struct Animation
{
	/++
	 + 初期化処理
	 +/
	static
	void startup()
	{
		Graphics g = Graphics.instance;
		renderer_ = new AnimationPartRenderer(g.D3DCompile, g.device, g.width, g.height);
	}

	/++
	 + 終了処理
	 +/
	static
	void shutdown()
	{
		delete renderer_;
	}

	/++
	 + 内部データを初期化する
	 +/
	void release()
	{
		counter_ = 0;
		currentFrame_ = 0;
		loopLength_ = 0;
		timeline_ = null;
		collisionCounts_ = null;
		collisionCurrentFrames_ = null;
	}

	/++
	 + 再生開始
	 +
	 + Params:
	 +  timeline = 再生データ
	 +  loopLength = ループ回数 負の値を指定すると無限ループ
	 +  restart = trueを指定すると再生中の場合でも最初から再生する
	 +/
	void play(AnimationTimeline* timeline, int loopLength, bool restart)
	{
		if (timeline is null)
		{
			counter_ = 0;
			return;
		}

		if (!restart && counter_ != 0 && timeline_ == timeline)
			return;

		// パラメータ設定
		counter_ = timeline.animationFrames[0].time;
		currentFrame_ = 0;
		loopLength_ = loopLength;
		timeline_ = timeline;

		uint collisionLength = timeline.collisionTimelines is null ? 0 : timeline.collisionTimelines.length;
		collisionCounts_.length = collisionLength;
		collisionCurrentFrames_.length = collisionLength;
		for (uint i; i < collisionLength; ++i)
		{
			collisionCounts_[i] = timeline.collisionTimelines[i].frames[0].time;
			collisionCurrentFrames_[i] = 0;
		}
	}

	/++
	 + 停止
	 +/
	void stop()
	{
		counter_ = 0;
	}

	/++
	 + 衝突判定データを取得
	 +
	 + Params:
	 +  index = インデックス
	 +/
	CollisionFrame*	getCurrentCollisionFrame(uint index)
	{
		if (timeline_ is null || timeline_.collisionTimelines is null || timeline_.collisionTimelines.length <= index)
			return null;
		else
			return &(timeline_.collisionTimelines[index].frames[collisionCurrentFrames_[index]]);
	}

	/++
	 + 毎フレームの処理
	 +
	 + アニメーションをさせるために毎フレーム必ず呼び出すこと
	 + これを呼び出さないとアニメーションは停止したままとなる
	 +/
	void move()
	{
		if (timeline_ is null)
			return;

		if (counter_ == 0)
			return;

		// アニメーションフレームの表示時間と比較
		if (--counter_ == 0)
		{
			// 新しいフレームに切り替える
			if (++currentFrame_ >= timeline_.length)
			{
				if (loopLength_ > 0)
					--loopLength_;

				if (loopLength_ != 0)
					currentFrame_ = 0;
				else
					--currentFrame_; // 最終フレームで固定
			}

			// カウンタを再設定
			counter_ = timeline_.animationFrames[currentFrame_].time;
		}

		uint collisionLength = timeline_.collisionTimelines is null ? 0 : timeline_.collisionTimelines.length;
		for (uint i; i < collisionLength; ++i)
		{
			if (collisionCounts_[i] != 0 && --collisionCounts_[i] == 0)
			{
				// 新しいフレームに切り替える
				if (++collisionCurrentFrames_[i] >= timeline_.collisionTimelines[i].frames.length)
				{
					if (loopLength_ != 0)
						collisionCurrentFrames_[i] = 0;
					else
						--collisionCurrentFrames_[i]; // 最終フレームで固定
				}

				// カウンタを再設定
				collisionCounts_[i] = timeline_.collisionTimelines[i].frames[collisionCurrentFrames_[i]].time;
			}
		}
	}

	/++
	 + 描画処理
	 +
	 + Params:
	 +  x = 描画位置X座標
	 +  y = 描画位置Y座標
	 +/
	void draw(float x, float y)
	{
		if (timeline_ !is null)
			timeline_.animationFrames[currentFrame_].draw(x, y);
	}

private:

	static
	AnimationPartRenderer renderer_;


	// 再生時間のカウンタ
	short counter_;

	// 表示中のフレーム
	int currentFrame_;

	// ループ回数 (負の数を指定すると無限ループ)
	int loopLength_;

	// 再生するアニメーション
	AnimationTimeline* timeline_;

	// 当り判定の再生時間カウンタ
	int[] collisionCounts_;

	// 当り判定の使用中フレーム
	int[] collisionCurrentFrames_;
}


/++
 + アニメーション表示処理
 +
 + どっかで初期化/終了せにゃならんので、Animation.startupとshutdownでセットする
 +/
class AnimationPartRenderer
{
private:

	// 定数バッファ
	static
	struct Constants
	{
		float[4] xarray;
		float[4] yarray;
		int u;             // Pattern.u
		int v;             // Pattern.v
		int width;         // Pattern.width
		int height;        // Pattern.height
		int textureWidth;  // Surface.width
		int textureHeight; // Surface.height
		int screenWidth;   // スクリーン幅の 1/2
		int screenHeight;  // スクリーン高さの 1/2
		float x;
		float y;
		uint color;        // Sprite.color
		uint dummy;        // 未使用(16byte境界に合わせないと定数バッファが作れない)
	}

	// 頂点フォーマット
	static
	struct Vertex
	{
		float x;
		float y;
		float u;
		float v;
		uint index;
	}

	// 頂点バッファ(システムメモリ)
	Vertex[4] vertices_;

	// 定数バッファ(システムメモリ)
	Constants constants_;

	// Input layout
	ID3D11InputLayout inputLayout_;

	// Vertex buffer
	ID3D11Buffer vertexBuffer_;

	// Vertex shader
	ID3D11VertexShader vertexShader_;

	// Vertex shader 定数バッファ
	ID3D11Buffer vertexShaderConstantsBuffer_;

	// Rasterizer設定
	ID3D11RasterizerState rasterizerState_;

	// Pixel shader
	ID3D11PixelShader pixelShader_;

	// Sampler
	ID3D11SamplerState samplerState_;

	// アルファブレンド設定
	ID3D11BlendState blendState_;

	/+
	 + コンストラクタ
	 +
	 + Params:
	 +  D3DCompile = 各種シェーダをコンパイルするための関数
	 +  device = ID3D11Device
	 +/
	this(D3DCompileFunction D3DCompile, ID3D11Device device, int screenWidth, int screenHeight)
	{
		constants_.screenWidth = screenWidth / 2;
		constants_.screenHeight = screenHeight /2;

		ID3DBlob compiled;
		ID3DBlob errors;

		//--------------------------------------------------------------
		// Input assembler ステージで使用するデータを準備

		// Vertex shader をコンパイル
		string vertexShaderSource = `
			cbuffer Constants : register(b0)
			{
				float4 xarray;
				float4 yarray;
				int4 pattern;
				int2 surface;
				int2 screen;
				float2 position;
				uint color;
			};

			struct VS_INPUT
			{
				float4 pos : POSITION;
				float2 uv : TEXCOORD;
				uint index : BLENDINDICES;
			};

			struct VS_OUTPUT
			{
				float4 pos : SV_POSITION;
				float4 col : COLOR;
				float2 uv : TEXCOORD;
			};

			VS_OUTPUT vertexShader(VS_INPUT input)
			{
				VS_OUTPUT output;
				output.pos.x = xarray[input.index];
				output.pos.y = yarray[input.index];
				output.pos.z = 0;
				output.pos.w = 1;

				output.pos.xy += position;

				// ワールド座標 → スクリーン座標変換
				output.pos.xy -= screen;
				output.pos.xy /= screen;
				output.pos.y = -output.pos.y;

				// テクスチャのUV座標計算
				input.uv *= pattern.zw;
				input.uv += pattern.xy;
				output.uv = input.uv / surface;

				// 色をfloat4に変換
				output.col.x = (color >> 16) & 0xff;
				output.col.y = (color >> 8) & 0xff;
				output.col.z = color & 0xff;
				output.col.w = (color >> 24) & 0xff;
				output.col /= 255;

				return output;
			}
		`;
		if (FAILED(D3DCompile(vertexShaderSource.ptr, vertexShaderSource.length, null, null, null, "vertexShader\0".ptr, "vs_4_0\0".ptr, 0, 0, &compiled, &errors)))
		{
			string message = "";
			if (errors !is null)
			{
				message = (cast(char*)errors.GetBufferPointer())[0 .. errors.GetBufferSize()].idup;
				errors.Release();
			}
			throw new GameError("Failed compile vertex shader: " ~ message);
		}

		// 入力レイアウトを作成
		D3D11_INPUT_ELEMENT_DESC[3] inputElementDesc;
		inputElementDesc[0] = D3D11_INPUT_ELEMENT_DESC("POSITION\0".ptr, 0, DXGI_FORMAT.DXGI_FORMAT_R32G32B32A32_FLOAT, 0, 0, D3D11_INPUT_CLASSIFICATION.D3D11_INPUT_PER_VERTEX_DATA, 0);
		inputElementDesc[1] = D3D11_INPUT_ELEMENT_DESC("TEXCOORD\0".ptr, 0, DXGI_FORMAT.DXGI_FORMAT_R32G32_FLOAT, 0, 8, D3D11_INPUT_CLASSIFICATION.D3D11_INPUT_PER_VERTEX_DATA, 0);
		inputElementDesc[2] = D3D11_INPUT_ELEMENT_DESC("BLENDINDICES\0".ptr, 0, DXGI_FORMAT.DXGI_FORMAT_R32_UINT, 0, 16, D3D11_INPUT_CLASSIFICATION.D3D11_INPUT_PER_VERTEX_DATA, 0);
		if (FAILED(device.CreateInputLayout(inputElementDesc.ptr, inputElementDesc.length, compiled.GetBufferPointer(), compiled.GetBufferSize(), &inputLayout_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();

			release();
			throw new GameError("Failed create input layout");
		}

		//--------------------------------------------------------------
		// Vertex Shader ステージで使用するデータを準備

		// Vertex shaderを作成
		if (FAILED(device.CreateVertexShader(compiled.GetBufferPointer(), compiled.GetBufferSize(), null, &vertexShader_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();
			release();
			throw new GameError("Failed create vertex shader");
		}

		compiled.Release();
		if (errors !is null)
			errors.Release();

		// 頂点データ
		vertices_[0] = Vertex(0, 0, 0, 0, 0);
		vertices_[1] = Vertex(0, 0, 1, 0, 1);
		vertices_[2] = Vertex(0, 0, 0, 1, 2);
		vertices_[3] = Vertex(0, 0, 1, 1, 3);

		D3D11_BUFFER_DESC vertexBufferDesc;
		vertexBufferDesc.ByteWidth = Vertex.sizeof * vertices_.length;
		vertexBufferDesc.Usage = D3D11_USAGE.D3D11_USAGE_DEFAULT;
		vertexBufferDesc.BindFlags = D3D11_BIND_FLAG.D3D11_BIND_VERTEX_BUFFER;
		vertexBufferDesc.CPUAccessFlags = 0;
		vertexBufferDesc.MiscFlags = 0;

		D3D11_SUBRESOURCE_DATA vertexBufferSubresource;
		vertexBufferSubresource.pSysMem = vertices_.ptr;
		if (FAILED(device.CreateBuffer(&vertexBufferDesc, &vertexBufferSubresource, &vertexBuffer_)))
		{
			release();
			throw new GameError("Failed create vertex buffer");
		}

		// 頂点シェーダ定数バッファを作成
		D3D11_BUFFER_DESC vertexShaderConstantsBufferDesc = D3D11_BUFFER_DESC(Constants.sizeof, D3D11_USAGE.D3D11_USAGE_DEFAULT, D3D11_BIND_FLAG.D3D11_BIND_CONSTANT_BUFFER, 0, 0);
		if (FAILED(device.CreateBuffer(&vertexShaderConstantsBufferDesc, null, &vertexShaderConstantsBuffer_)))
		{
			release();
			throw new GameError("Failed create constants buffer");
		}

		//--------------------------------------------------------------
		// Rasterizer ステージで使用するデータを準備


		// Rasterizer state
		D3D11_RASTERIZER_DESC rasterizerDesc;
		rasterizerDesc.FillMode = D3D11_FILL_MODE.D3D11_FILL_SOLID;
		rasterizerDesc.CullMode = D3D11_CULL_MODE.D3D11_CULL_NONE;
		rasterizerDesc.FrontCounterClockwise = FALSE;
		rasterizerDesc.DepthBias = 0;
		rasterizerDesc.DepthBiasClamp = 0;
		rasterizerDesc.SlopeScaledDepthBias = 0;
		rasterizerDesc.DepthClipEnable = FALSE;
		rasterizerDesc.ScissorEnable = FALSE;
		rasterizerDesc.MultisampleEnable = FALSE;
		rasterizerDesc.AntialiasedLineEnable = TRUE;
		if (FAILED(device.CreateRasterizerState(&rasterizerDesc, &rasterizerState_)))
		{
			release();
			throw new GameError("Failed create rasterizer state");
		}

		//--------------------------------------------------------------
		// Pixel Shader ステージで使用するデータを準備

		// Pixel shader
		string pixelShaderSource =
		`
			struct PS_INPUT
			{
				float4 pos : SV_POSITION;
				float4 col : COLOR;
				float2 uv : TEXCOORD;
			};

			Texture2D tex : register(t0);

			SamplerState sam : register(s0);

			float4 pixelShader(PS_INPUT input) : SV_Target
			{
				return tex.Sample(sam, input.uv) * input.col;
			}
		`;
		if (FAILED(D3DCompile(pixelShaderSource.ptr, pixelShaderSource.length, null, null, null, "pixelShader\0".ptr, "ps_4_0\0".ptr, 0, 0, &compiled, &errors)))
		{
			string message = "";
			if (errors !is null)
			{
				message = (cast(char*)errors.GetBufferPointer())[0 .. errors.GetBufferSize()].idup;
				errors.Release();
			}

			throw new GameError("Failed compile pixel shader: " ~ message);
		}

		// Pixel shaderを作成
		if (FAILED(device.CreatePixelShader(compiled.GetBufferPointer(), compiled.GetBufferSize(), null, &pixelShader_)))
		{
			compiled.Release();
			if (errors !is null)
				errors.Release();
			release();
			throw new GameError("Failed create vertex shader");
		}

		compiled.Release();
		if (errors !is null)
			errors.Release();


		D3D11_SAMPLER_DESC samplerDesc;
		samplerDesc.Filter = D3D11_FILTER.D3D11_FILTER_ANISOTROPIC;
		samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_MODE.D3D11_TEXTURE_ADDRESS_CLAMP;
		samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_MODE.D3D11_TEXTURE_ADDRESS_CLAMP;
		samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_MODE.D3D11_TEXTURE_ADDRESS_CLAMP;
		samplerDesc.MipLODBias = 0.0f;
		samplerDesc.MaxAnisotropy = 8;
		samplerDesc.ComparisonFunc = D3D11_COMPARISON_FUNC.D3D11_COMPARISON_ALWAYS;
		samplerDesc.BorderColor[0] = 0.0f;
		samplerDesc.BorderColor[1] = 0.0f;
		samplerDesc.BorderColor[2] = 0.0f;
		samplerDesc.BorderColor[3] = 0.0f;
		samplerDesc.MinLOD = 0.0f;
		samplerDesc.MaxLOD = D3D11_FLOAT32_MAX;
		if (FAILED(device.CreateSamplerState(&samplerDesc, &samplerState_)))
		{
			release();
		    throw new GameError("Failed create clamp sampler state");
		}

		//--------------------------------------------------------------
		// Output Merger ステージで使用するデータを準備

		// アルファブレンドの設定
		D3D11_BLEND_DESC blendStateDesc;
		blendStateDesc.AlphaToCoverageEnable = FALSE;
		blendStateDesc.IndependentBlendEnable = FALSE;
		blendStateDesc.RenderTarget[0].BlendEnable = TRUE;
		blendStateDesc.RenderTarget[0].SrcBlend = D3D11_BLEND.D3D11_BLEND_SRC_ALPHA;
		blendStateDesc.RenderTarget[0].DestBlend = D3D11_BLEND.D3D11_BLEND_INV_SRC_ALPHA;
		blendStateDesc.RenderTarget[0].BlendOp = D3D11_BLEND_OP.D3D11_BLEND_OP_ADD;
		blendStateDesc.RenderTarget[0].SrcBlendAlpha = D3D11_BLEND.D3D11_BLEND_ZERO;
		blendStateDesc.RenderTarget[0].DestBlendAlpha = D3D11_BLEND.D3D11_BLEND_ONE;
		blendStateDesc.RenderTarget[0].BlendOpAlpha = D3D11_BLEND_OP.D3D11_BLEND_OP_ADD;
		blendStateDesc.RenderTarget[0].RenderTargetWriteMask = D3D11_COLOR_WRITE_ENABLE.D3D11_COLOR_WRITE_ENABLE_ALL;
		if (FAILED(device.CreateBlendState(&blendStateDesc, &blendState_)))
		{
			release();
			throw new GameError("Failed create blend state");
		}
	}

	/+
	 + デストラクタ
	 +/
	~this()
	{
		release();
	}

	/+
	 + 描画処理
	 +
	 + 呼び出す前にConstantsにデータを詰めておく
	 +/
	void render(ID3D11DeviceContext context, ID3D11ShaderResourceView textureResourceView)
	{
		// Input Assember ステージに入力するデータ形式を設定
		context.IASetInputLayout(inputLayout_);

		// Input Assember ステージに入力されたデータの使用方法を設定
		context.IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY.D3D11_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP);

		// Input Assembler ステージに頂点をセット
		uint stride = Vertex.sizeof;
		uint offset = 0;

		// 頂点バッファをセット
		context.IASetVertexBuffers(0, 1, &vertexBuffer_, &stride, &offset);

		// Vertex Shader ステージに Vertex Shader を設定
		context.VSSetShader(vertexShader_, null, 0);

		// 定数バッファの内容を更新
		context.UpdateSubresource(vertexShaderConstantsBuffer_, 0, null, &constants_, 0, 0);

		// Vertex Shaderに定数バッファをセット
		context.VSSetConstantBuffers(0, 1, &vertexShaderConstantsBuffer_);

		// Rasterizer ステージの設定
		context.RSSetState(rasterizerState_);

		// Pixel Shaderに Pixel Shader を設定
		context.PSSetShader(pixelShader_, null, 0);

		// テクスチャをセット
		context.PSSetShaderResources(0, 1, &textureResourceView);

		// Pixel Shaderに Sampler を設定
		context.PSSetSamplers(0, 1, &samplerState_);

		// Output Merger ステージでアルファブレンドを設定
		context.OMSetBlendState(blendState_, null, 0xffffffff);


		// パイプラインを実行して描画する
		context.Draw(vertices_.length, 0);
	}

	/+
	 + リソースの解放処理
	 +/
	void release()
	{
		if (blendState_ !is null)
		{
			blendState_.Release();
			blendState_ = null;
		}

		if (samplerState_ !is null)
		{
			samplerState_.Release();
			samplerState_ = null;
		}

		if (pixelShader_ !is null)
		{
			pixelShader_.Release();
			pixelShader_ = null;
		}

		if (rasterizerState_ !is null)
		{
			rasterizerState_.Release();
			rasterizerState_ = null;
		}

		if (vertexShaderConstantsBuffer_ !is null)
		{
			vertexShaderConstantsBuffer_.Release();
			vertexShaderConstantsBuffer_ = null;
		}

		if (vertexShader_ !is null)
		{
			vertexShader_.Release();
			vertexShader_ = null;
		}

		if (vertexBuffer_ !is null)
		{
			vertexBuffer_.Release();
			vertexBuffer_ = null;
		}

		if (inputLayout_ !is null)
		{
			inputLayout_.Release();
			inputLayout_ = null;
		}
	}
}

