/++
 + Vorbis file access header module for D programming language
 +
 + Use, distribution and reproduction of this library source is 
 + governed by a BSD-STYLE souce license included with this source in 'COPYING'.
 + Please read these terms before distributing.
 +
 + Copyright(C) 2009-2012 Takuya Kurosawa
 +
 + Original code:
 +  vorbisfile.h (libvorbis-1.3.2)
 +  THE OggVorbis SOURCE CODE IS (C) COPYRIGHT 1994-2007
 +  by the Xiph.Org Foundation http://www.xiph.org/
 +/
module org.xiph.libvorbis.vorbisfile;

import core.stdc.stdio;
import org.xiph.libogg.types;
import org.xiph.libogg.ogg;
import org.xiph.libvorbis.codec;

/++
 + The function prototypes for the callbacks are basically the same as for
 + the stdio functions fread, fseek, fclose, ftell.
 + The one difference is that the FILE* arguments have been replaced with
 + a void* - this is to be used as a pointer to whatever internal data these
 + functions might need. In the stdio case, it's just a FILE* cast to a void*
 +
 + If you use other functions, check the docs for these functions and return
 + the right values. For seek_func(), you *MUST* return -1 if the stream is unseekable
 +/
extern (C)
{
	alias size_t function(void* ptr, size_t size, size_t nmemb, void* datasource) ov_callback_read_func;
	alias int function(void* datasource, ogg_int64_t offset, int whence) ov_callback_seek_func;
	alias int function(void* datasource) ov_callback_close_func;
	alias int function(void* datasource) ov_callback_tell_func;
}

struct ov_callbacks
{
	ov_callback_read_func read_func;
	ov_callback_seek_func seek_func;
	ov_callback_close_func close_func;
	ov_callback_tell_func tell_func;
}

int _ov_header_fseek_wrap(FILE* f, ogg_int64_t off, int whence)
{
	if (f is null)
		return -1;

	return core.stdc.stdio.fseek(f, cast(int)off, whence);
}

/+
 + These structs below (OV_CALLBACKS_DEFAULT etc) are defined here as static data.
 + That means that every file which import this header module will get its own copy of these structs.
 + These static symbols are essential on platforms such as Windows on
 + which several different versions of stdio support may be linked to
 + by different DLLs, and we need to be certain we know which one
 + we're using (the same one as the main application).
 +/

ov_callbacks OV_CALLBACKS_DEFAULT =
{
	cast(ov_callback_read_func)&core.stdc.stdio.fread,
	cast(ov_callback_seek_func)&_ov_header_fseek_wrap,
	cast(ov_callback_close_func)&core.stdc.stdio.fclose,
	cast(ov_callback_tell_func)&core.stdc.stdio.ftell
};

ov_callbacks OV_CALLBACKS_NOCLOSE =
{
	cast(ov_callback_read_func)&core.stdc.stdio.fread,
	cast(ov_callback_seek_func)&_ov_header_fseek_wrap,
	null,
	cast(ov_callback_tell_func)&core.stdc.stdio.ftell
};

ov_callbacks OV_CALLBACKS_STREAMONLY =
{
	cast(ov_callback_read_func)&core.stdc.stdio.fread,
	null,
	cast(ov_callback_close_func)&core.stdc.stdio.fclose,
	null
};

ov_callbacks OV_CALLBACKS_STREAMONLY_NOCLOSE =
{
	cast(ov_callback_read_func)&core.stdc.stdio.fread,
	null,
	null,
	null
};

immutable int NOTOPEN   = 0;
immutable int PARTOPEN  = 1;
immutable int OPENED    = 2;
immutable int STREAMSET = 3;
immutable int INITSET   = 4;

struct OggVorbis_File
{
	// Pointer to a FILE *, etc.
	void* datasource;

	int seekable;
	ogg_int64_t offset;
	ogg_int64_t end;
	ogg_sync_state oy;

	// If the FILE handle isn't seekable (eg, a pipe), only the current stream appears
	int links;
	ogg_int64_t* offsets;
	ogg_int64_t* dataoffsets;
	int* serialnos;

	// overloaded to maintain binary compatibility; x2 size, stores both beginning and end values
	ogg_int64_t* pcmlengths;

	vorbis_info* vi;
	vorbis_comment* vc;

	// Decoding working state local storage
	ogg_int64_t pcm_offset;
	int ready_state;
	int current_serialno;
	int current_link;

	double bittrack;
	double samptrack;

	// take physical pages, weld into a logical stream of packets
	ogg_stream_state os;

	// central working state for the packet->PCM decoder
	vorbis_dsp_state vd;

	// local working space for packet->PCM decode
	vorbis_block vb;

	ov_callbacks callbacks;
}

extern (C)
{
	int ov_clear(OggVorbis_File* vf);
	int ov_fopen(const(char)* path, OggVorbis_File* vf);
	int ov_open(FILE* f, OggVorbis_File* vf, const(char)* initial, int ibytes);
	int ov_open_callbacks(void* datasource, OggVorbis_File* vf, const(char)* initial, int ibytes, ov_callbacks callbacks);
	int ov_test(FILE* f, OggVorbis_File* vf, const(char)* initial, int ibytes);
	int ov_test_callbacks(void* datasource, OggVorbis_File* vf, const(char)* initial, int ibytes, ov_callbacks callbacks);
	int ov_test_open(OggVorbis_File* vf);

	int ov_bitrate(OggVorbis_File* vf, int i);
	int ov_bitrate_instant(OggVorbis_File* vf);
	int ov_streams(OggVorbis_File* vf);
	int ov_seekable(OggVorbis_File* vf);
	int ov_serialnumber(OggVorbis_File* vf, int i);

	ogg_int64_t ov_raw_total(OggVorbis_File* vf, int i);
	ogg_int64_t ov_pcm_total(OggVorbis_File* vf, int i);
	double ov_time_total(OggVorbis_File* vf, int i);

	int ov_raw_seek(OggVorbis_File* vf, ogg_int64_t pos);
	int ov_pcm_seek(OggVorbis_File* vf, ogg_int64_t pos);
	int ov_pcm_seek_page(OggVorbis_File* vf, ogg_int64_t pos);
	int ov_time_seek(OggVorbis_File* vf, double pos);
	int ov_time_seek_page(OggVorbis_File* vf, double pos);

	int ov_raw_seek_lap(OggVorbis_File* vf, ogg_int64_t pos);
	int ov_pcm_seek_lap(OggVorbis_File* vf, ogg_int64_t pos);
	int ov_pcm_seek_page_lap(OggVorbis_File* vf, ogg_int64_t pos);
	int ov_time_seek_lap(OggVorbis_File* vf, double pos);
	int ov_time_seek_page_lap(OggVorbis_File* vf, double pos);

	ogg_int64_t ov_raw_tell(OggVorbis_File* vf);
	ogg_int64_t ov_pcm_tell(OggVorbis_File* vf);
	double ov_time_tell(OggVorbis_File* vf);

	vorbis_info* ov_info(OggVorbis_File* vf, int link);
	vorbis_comment* ov_comment(OggVorbis_File* vf, int link);

	int ov_read_float(OggVorbis_File* vf, float*** pcm_channels,int samples, int* bitstream);
	int ov_read_filter(OggVorbis_File* vf, char* buffer, int length, int bigendianp, int word, int sgned, int* bitstream, void function (float** pcm, int channels, int samples, void* filter_param) filter, void* filter_param);
	int ov_read(OggVorbis_File* vf, char* buffer, int length, int bigendianp, int word, int sgned, int* bitstream);
	int ov_crosslap(OggVorbis_File* vf1,OggVorbis_File* vf2);

	int ov_halfrate(OggVorbis_File* vf, int flag);
	int ov_halfrate_p(OggVorbis_File* vf);
}